﻿
namespace KPGBase.Excercises
{
    internal class Excercise05 : BasicExcercise
    {
        // click
        private const float MIN_DISTANCE = 5f; 
        int selectedPoint;       

        // Bezier curve
        /// <summary> Number of points of the control polygon for one bezier arc </summary>
        int order = 4; 
        /// <summary> Points of the control polygon </summary>
        List<PointF> points;              
        
        // DeCasteljau
        /// <summary> Parameter t for deCasteljau </summary>
        float t = 0.3f;
        /// <summary> Points of the line segments created using deCasteljau </summary>
        List<PointF> deCasteljau;  

        public Excercise05(Bitmap canvas) : base(canvas)
        {
            // default testing points
            points = new List<PointF>()
            {
                new PointF(50, 50),
                new PointF(300, 50),
                new PointF(300, 300),
                new PointF(550, 300),
            };
        }

        /// <summary>
        /// Clear all points from control polygon
        /// </summary>
        internal void ClearPoints()
        {
            points = new List<PointF>();
            deCasteljau = new List<PointF>();
            DrawBezierCurve();
        }
        
        /// <summary>
        /// Draw Bezier curve
        /// </summary>
        public void DrawBezierCurve()
        {
            ClearImage();
            DrawControlPolygon();

            DeCasteljau(t);
            DrawDeCasteljau();

            DrawBezier();
        }

        #region Bezier curve basis function

        /// <summary>
        /// Factorial using iteration method
        /// </summary>
        private int Factorial(int x)
        {
            int fac = 1;
            for (int i = x; i > 1; i--)
            {
                fac = i * fac;
            }
            return fac;
        }

        /// <summary>
        /// Combination number
        /// </summary>
        private double CombNum(int n, int i)
        {
            double ni;
            double a1 = Factorial(n);
            double a2 = Factorial(i);
            double a3 = Factorial(n - i);
            ni = a1 / (a2 * a3);
            return ni;
        }

        /// <summary>
        /// Bernstein polynom
        /// </summary>
        private double Bernstein(int n, int i, double t)
        {
            double basis;
            double ti;  // t^i
            double tni; // (1 - t)^i

            // Prevent problems with pow
            if (t == 0.0 && i == 0)
                ti = 1.0;
            else
                ti = Math.Pow(t, i);

            if (n == i && t == 1.0)
                tni = 1.0;
            else
                tni = Math.Pow((1 - t), (n - i));

            // Bernstein basis
            basis = CombNum(n, i) * ti * tni;
            return basis;
        }

        #endregion

        /// <summary>
        /// Draw control polygon - points saved in attribute points
        /// </summary>
        private void DrawControlPolygon()
        {
            float r = MIN_DISTANCE;
            for (int i = 0; i < points.Count; i++)
                g.FillEllipse(Brushes.Pink, points[i].X - r, points[i].Y - r, 2 * r, 2 * r);

            Pen pen = new Pen(Color.LightGray, 2);
            for (int i = 1; i < points.Count; i++)
                g.DrawLine(pen, points[i - 1].X, points[i - 1].Y, points[i].X, points[i].Y);
        }

        #region Bezier curve

        /// <summary>
        /// Draw bezier curve on the given input points in attribute points
        /// </summary>
        private void DrawBezier()
        {
            // number of bezier arcs
            int arcCount = (points.Count - 1) / (order - 1);
            // number of points on one arc
            const int POINTS_ON_CURVE = 10; 

            for (int i = 0; i < arcCount; i++)
            {
                // control points for the given arc
                PointF[] affectedPoints = new PointF[order];
                // copy - array from, starting index, array to, indices to copy to
                Array.Copy(points.ToArray(), i * (order - 1), affectedPoints, 0, order);

                // Bezier arc
                PointF[] p = Bezier2D(affectedPoints, POINTS_ON_CURVE);

                // draw the arc
                Pen pen = new Pen(Color.White, 2);
                g.DrawLines(pen, p);
            }
        }

        /// <summary>
        /// Bezier curve on points b
        /// </summary>
        /// <param name="b"> Control points of the bezier curve </param>
        /// <param name="ptsOut"> Required number of created points on the bezier curve </param>
        /// <returns> Points on the bezier curve </returns>
        public PointF[] Bezier2D(PointF[] b, int ptsOut)
        {
            // Created points
            PointF[] p = new PointF[ptsOut];

            // TODO
            // Create a bezier curve
            //  - in array b are the control points
            //  - create exactly ptsOut points on the bezier curve
            //  - test for different parameter ptsOut

            // number of control points
            int npts = b.Length;
            // current t
            double t = 0;
            // t step parameter
            double step = 1.0 / (ptsOut - 1);

            for (int n = 0; n != ptsOut; n++)
            {
                p[n] = new Point(0, 0);
                for (int i = 0; i != npts; i++)
                {
                    double basis = Bernstein(npts - 1, i, t);
                    p[n] = new PointF((float)(p[n].X + basis * b[i].X), (float)(p[n].Y + basis * b[i].Y));
                }

                t += step;
            }

            return p;
        }

        #endregion

        #region DeCasteljau

        /// <summary>
        /// Draw deCasteljau steps - points saved in attribute deCasteljau
        /// </summary>
        private void DrawDeCasteljau()
        {
            if (deCasteljau == null)
                return;

            Pen pen = new Pen(Color.LightGray, 2f);
            for (int i = 1; i < deCasteljau.Count; i = i + 2)
            {
                g.DrawLine(pen, deCasteljau[i - 1].X, deCasteljau[i - 1].Y, deCasteljau[i].X, deCasteljau[i].Y);
            }
        }

        /// <summary>
        /// Compute deCasteljau line segments for given parameter t
        /// </summary>
        /// <param name="t"> Parameter t </param>
        private void DeCasteljau(float t)
        {
            deCasteljau = new List<PointF>();
            // number of arcs from given control points
            int arcCount = (points.Count - 1) / (order - 1);

            // is there enough points for a bezier curve
            if (points.Count < order)
                return;
            
            // for each arc
            for (int i = 0; i < arcCount; i++)
            {
                // arc control points
                PointF[] arcPts = new PointF[order];
                Array.Copy(points.ToArray(), i * (order - 1), arcPts, 0, order);

                // Compute deCasteljau line segments
                //  - find out how many line segments will be needed for the order of the curve
                //  - for each arc compute all line segments (for each control line will be needed one dividing point

                // how many line segments to create
                int temp = order - 2;
                // until deCasteljau steps reach one point
                while (temp > 0)
                {
                    // new line segment points
                    PointF[] newPoints = new PointF[arcPts.Length - 1];
                    for (int j = 0; j < temp; j++)
                    {
                        // create a line segment from those points
                        PointF a = arcPts[j];
                        PointF b = arcPts[j + 1];
                        PointF c = arcPts[j + 2];

                        // new line segment
                        newPoints[j] = new PointF(a.X + t * (b.X - a.X), a.Y + t * (b.Y - a.Y));
                        newPoints[j + 1] = new PointF(b.X + t * (c.X - b.X), b.Y + t * (c.Y - b.Y));
                        deCasteljau.Add(newPoints[j]);
                        deCasteljau.Add(newPoints[j + 1]);
                    }
                    // next step of deCasteljau uses the newly created "control polygon"
                    arcPts = newPoints;
                    // one less line segment in next iteration
                    temp--;
                }
            }
        }
        #endregion

        #region Point selection

        /// <summary>
        /// Select a control point or add a new one
        /// </summary>
        /// <param name="point"> Input coordinates </param>
        internal void SelectPoint(Point point)
        {
            // Point is selected if any control point is closer than MIN_DISTANCE
            // New control point is added otherwise
            // Attributes
            //  - selectedPoint - index of point that the user clicked on
            // Selected point can now be dragged while left button remains clicked

            // find if any point close enough
            float x, y;
            bool found = false;
            for (int i = 0; i < points.Count; i++)
            {
                x = point.X - points[i].X;
                y = point.Y - points[i].Y;

                // clicked close enough
                if (x * x + y * y < MIN_DISTANCE * MIN_DISTANCE)
                {
                    found = true;
                    selectedPoint = i;
                    break;
                }
            }

            // no selected point
            if (!found)
            {
                points.Add(new Point(point.X, point.Y));
                selectedPoint = points.Count - 1;
                DrawBezierCurve();
            }
        }

        /// <summary>
        /// Move selected point on canvas
        /// </summary>
        internal void MoveSelectedPoint(Point newLoc)
        {
            // if a point is selected then move it to new location
            points[selectedPoint] = newLoc;
            DrawBezierCurve();
        }

        #endregion
    }
}
